""" Utility functions and classes for the protocol api """
from __future__ import annotations

from collections import UserDict
import functools
import logging
from dataclasses import dataclass, field, astuple
from typing import (Any, Callable, Dict, Optional,
                    TYPE_CHECKING, Union, List)

from opentrons import types as top_types
from opentrons.protocols.api_support.types import APIVersion
from opentrons.hardware_control import (types, SynchronousAdapter, API,
                                        HardwareAPILike, ThreadManager)
if TYPE_CHECKING:
    from opentrons.protocols.context.instrument import \
        AbstractInstrument
    from opentrons.protocol_api.labware import Well, Labware
    from opentrons.protocols.geometry.deck import Deck
    from opentrons.hardware_control.dev_types import HasLoop

MODULE_LOG = logging.getLogger(__name__)


class APIVersionError(Exception):
    """
    Error raised when a protocol attempts to access behavior not implemented
    """
    pass


def _assert_gzero(val: Any, message: str) -> float:
    try:
        new_val = float(val)
        assert new_val > 0.0
        return new_val
    except (TypeError, ValueError, AssertionError):
        raise AssertionError(message)


@dataclass
class EdgeList:
    right: Optional[top_types.Point] = field(default_factory=top_types.Point)
    left: Optional[top_types.Point] = field(default_factory=top_types.Point)
    center: Optional[top_types.Point] = field(default_factory=top_types.Point)
    up: top_types.Point = field(default_factory=top_types.Point)
    down: top_types.Point = field(default_factory=top_types.Point)


def determine_edge_path(
        where: 'Well', mount: top_types.Mount,
        default_edges: EdgeList, deck: 'Deck') -> EdgeList:
    left_path = EdgeList(
        left=default_edges.left,
        right=None,
        center=default_edges.center,
        up=default_edges.up,
        down=default_edges.down)
    right_path = EdgeList(
        left=None,
        right=default_edges.right,
        center=default_edges.center,
        up=default_edges.up,
        down=default_edges.down
    )
    labware = where.parent

    r_mount = top_types.Mount.RIGHT
    l_mount = top_types.Mount.LEFT
    l_col = labware.columns()[0]
    r_col = labware.columns()[-1]
    right_pip_criteria = mount is r_mount and where in l_col
    left_pip_criteria = mount is l_mount and where in r_col

    next_to_mod = deck.is_edge_move_unsafe(mount, labware)
    if labware.parent in ['3', '6', '9'] and left_pip_criteria:
        return left_path
    elif left_pip_criteria and next_to_mod:
        return left_path
    elif right_pip_criteria and next_to_mod:
        return right_path
    return default_edges


def build_edges(
        where: 'Well', offset: float, mount: top_types.Mount,
        deck: 'Deck', radius: float = 1.0,
        version: APIVersion = APIVersion(2, 7)) -> List[top_types.Point]:
    # Determine the touch_tip edges/points
    offset_pt = top_types.Point(0, 0, offset)
    edge_list = EdgeList(
        right=where.from_center_cartesian(x=radius, y=0, z=1) + offset_pt,
        left=where.from_center_cartesian(x=-radius, y=0, z=1) + offset_pt,
        center=where.from_center_cartesian(x=0, y=0, z=1) + offset_pt,
        up=where.from_center_cartesian(x=0, y=radius, z=1) + offset_pt,
        down=where.from_center_cartesian(x=0, y=-radius, z=1) + offset_pt
    )

    if version < APIVersion(2, 4):
        edge_list.center = None
        # Add the center value before switching axes
        return [edge for edge in astuple(edge_list) if edge]
    new_edges = determine_edge_path(where, mount, edge_list, deck)
    return [edge for edge in astuple(new_edges) if edge]


def labware_column_shift(
        initial_well: 'Well', tiprack: 'Labware',
        well_spacing: int = 4,
        modulo_value: int = 8) -> 'Well':
    unshifted_index = tiprack.wells().index(initial_well)
    unshifted_column = unshifted_index // modulo_value
    shifted_column = unshifted_column + well_spacing
    shifted_well = unshifted_index % modulo_value
    return tiprack.columns()[shifted_column][shifted_well]


class FlowRates:
    """ Utility class for rich setters/getters for flow rates """

    def __init__(self,
                 instr: AbstractInstrument) -> None:
        self._instr = instr

    def set_defaults(self, api_level: APIVersion):
        pipette = self._instr.get_pipette()
        self.aspirate = _find_value_for_api_version(
            api_level, pipette['default_aspirate_flow_rates'])
        self.dispense = _find_value_for_api_version(
            api_level, pipette['default_dispense_flow_rates'])
        self.blow_out = _find_value_for_api_version(
            api_level, pipette['default_blow_out_flow_rates'])

    @property
    def aspirate(self) -> float:
        return self._instr.get_pipette()['aspirate_flow_rate']

    @aspirate.setter
    def aspirate(self, new_val: float):
        self._instr.set_flow_rate(
            aspirate=_assert_gzero(
                new_val, 'flow rate should be a numerical value in ul/s'))

    @property
    def dispense(self) -> float:
        return self._instr.get_pipette()['dispense_flow_rate']

    @dispense.setter
    def dispense(self, new_val: float):
        self._instr.set_flow_rate(
            dispense=_assert_gzero(
                new_val, 'flow rate should be a numerical value in ul/s'))

    @property
    def blow_out(self) -> float:
        return self._instr.get_pipette()['blow_out_flow_rate']

    @blow_out.setter
    def blow_out(self, new_val: float):
        self._instr.set_flow_rate(
            blow_out=_assert_gzero(
                new_val, 'flow rate should be a numerical value in ul/s'))


def _find_value_for_api_version(for_version: APIVersion,
                                values: Dict[str, float]) -> float:
    """
    Parse a dict that looks like
    {"2.0": 5,
    "2.5": 4}
    (aka the flow rate values from pipette config) and return the value for
    the highest api level that is at or underneath ``for_version``
    """
    sorted_versions = sorted({APIVersion.from_string(k): v
                              for k, v in values.items()})
    last = values[str(sorted_versions[0])]
    for version in sorted_versions:
        if version > for_version:
            break
        last = values[str(version)]
    return last


class PlungerSpeeds:
    """ Utility class for rich setters/getters for speeds """

    def __init__(self,
                 instr: AbstractInstrument) -> None:
        self._instr = instr

    @property
    def aspirate(self) -> float:
        return self._instr.get_pipette()['aspirate_speed']

    @aspirate.setter
    def aspirate(self, new_val: float):
        self._instr.set_pipette_speed(
            aspirate=_assert_gzero(
                new_val, 'speed should be a numerical value in mm/s'))

    @property
    def dispense(self) -> float:
        return self._instr.get_pipette()['dispense_speed']

    @dispense.setter
    def dispense(self, new_val: float):
        self._instr.set_pipette_speed(
            dispense=_assert_gzero(
                new_val, 'speed should be a numerical value in mm/s'))

    @property
    def blow_out(self) -> float:
        return self._instr.get_pipette()['blow_out_speed']

    @blow_out.setter
    def blow_out(self, new_val: float):
        self._instr.set_pipette_speed(
            blow_out=_assert_gzero(
                new_val, 'speed should be a numerical value in mm/s'))


class Clearances:
    def __init__(self,
                 default_aspirate: float,
                 default_dispense: float) -> None:
        self._aspirate = default_aspirate
        self._dispense = default_dispense

    @property
    def aspirate(self) -> float:
        return self._aspirate

    @aspirate.setter
    def aspirate(self, new_val: float):
        self._aspirate = float(new_val)

    @property
    def dispense(self) -> float:
        return self._dispense

    @dispense.setter
    def dispense(self, new_val: float):
        self._dispense = float(new_val)


class AxisMaxSpeeds(UserDict):
    """ Special mapping allowing internal storage by Mount enums and
    user access by string
    """

    def __getitem__(self, key: Union[str, types.Axis]):
        checked_key = AxisMaxSpeeds._verify_key(key)
        return self.data[checked_key]

    @staticmethod
    def _verify_key(key: Any) -> types.Axis:
        if isinstance(key, types.Axis):
            checked_key: Optional[types.Axis] = key
        elif isinstance(key, str):
            checked_key = types.Axis[key.upper()]
        else:
            checked_key = None
        if checked_key not in types.Axis.gantry_axes():
            raise KeyError(key)
        return checked_key

    def __setitem__(self, key: Any, value: Any):
        if value is None:
            del self[key]
            return

        checked_key = AxisMaxSpeeds._verify_key(key)
        checked_val = _assert_gzero(
            value, 'max speeds should be numerical values in mm/s')

        self.data[checked_key] = checked_val

    def __delitem__(self, key: Union[str, types.Axis]):
        checked_key = AxisMaxSpeeds._verify_key(key)
        del self.data[checked_key]

    def __iter__(self):
        """ keys() and dict iteration return string keys """
        return (k.name for k in self.data.keys())

    def keys(self):
        return (k.name for k in self.data.keys())

    def items(self):
        return ((k.name, v) for k, v in self.data.items())


HardwareToManage = Union[ThreadManager,
                         SynchronousAdapter,
                         'HasLoop']


# TODO: BC 2020-03-02 This class's utility as a utility class is drying up.
# It's only job is to ensure that the hardware a given
# ProtocolContext references, is synchronously callable.
# All internal calls to ProtocolContext __init__
# or build_using, either pass a ThreadManaged API instance
# or pass None and expect HardwareManager to create one
# for them. It seems as though we could replace this
# with a single if else that covers just those two cases.
# If that were the case, perhaps it would be clearer to move
# this logic back into the ProtocolContext definition
# and hold onto a sync hardware api directly instead of
# through the ._hw_manager.hardware indirection.
class HardwareManager:
    def __init__(self, hardware: Optional[HardwareToManage]):
        if hardware is None:
            # TODO AL 20210209. This is a highly dangerous thread leak.
            self._current = ThreadManager(API.build_hardware_simulator).sync
        elif isinstance(hardware, SynchronousAdapter):
            self._current = hardware
        elif isinstance(hardware, ThreadManager):
            self._current = hardware.sync
        else:
            self._current = SynchronousAdapter(hardware)

    @property
    def hardware(self):
        return self._current

    def set_hw(self, hardware):
        if isinstance(hardware, SynchronousAdapter):
            self._current = hardware
        elif isinstance(hardware, ThreadManager):
            self._current = hardware.sync
        elif isinstance(hardware, HardwareAPILike):
            self._current = SynchronousAdapter(hardware)
        else:
            raise TypeError(
                "hardware should be API or synch adapter but is {}"
                .format(hardware))
        return self._current

    def reset_hw(self):
        # TODO AL 20210209. This is a highly dangerous thread leak.
        self._current = ThreadManager(API.build_hardware_simulator).sync
        return self._current


def clamp_value(
        input_value: float, max_value: float, min_value: float,
        log_tag: str = '') -> float:
    if input_value > max_value:
        MODULE_LOG.info(
            f'{log_tag} clamped input {input_value} to {max_value}')
        return max_value
    if input_value < min_value:
        MODULE_LOG.info(
            f'{log_tag} clamped input {input_value} to {min_value}')
        return min_value
    return input_value


def requires_version(
        major: int, minor: int) -> Callable[[Callable], Callable]:
    """ Decorator. Apply to Protocol API methods or attributes to indicate
    the first version in which the method or attribute was present.
    """
    def _set_version(decorated_obj: Callable) -> Callable:
        added_version = APIVersion(major, minor)
        setattr(decorated_obj, '__opentrons_version_added',
                added_version)
        if hasattr(decorated_obj, '__doc__'):
            # Add the versionadded stanza to everything decorated if we can
            docstr = decorated_obj.__doc__ or ''
            # this newline and initial space has to be there for sphinx to
            # parse this correctly and not add it into for instance a
            # previous code-block
            docstr += f'\n\n        .. versionadded:: {added_version}\n\n'
            decorated_obj.__doc__ = docstr

        @functools.wraps(decorated_obj)
        def _check_version_wrapper(*args, **kwargs):
            slf = args[0]
            added_in = decorated_obj.__opentrons_version_added  # type: ignore
            current_version = slf._api_version

            if APIVersion(2, 0) <= current_version < added_in:
                # __qualname__ is *probably* set on every kind of object we care
                # about, but the docs leave it ambiguous, so fall back to str().
                name = getattr(decorated_obj, "__qualname__",
                               str(decorated_obj))

                raise APIVersionError(
                    f'{name} was added in {added_in}, but your '
                    f'protocol requested version {current_version}. You '
                    f'must increase your API version to {added_in} to '
                    'use this functionality.')
            return decorated_obj(*args, **kwargs)

        return _check_version_wrapper

    return _set_version


class ModifiedList(list):
    def __contains__(self, item):
        for name in self:
            if name == item.replace("-", "_").lower():
                return True
        return False



from __future__ import annotations
import asyncio
from contextlib import contextmanager, ExitStack
import logging
from typing import (Any, Dict, List, Optional, Tuple,
                    TYPE_CHECKING, Union, Sequence)
from typing_extensions import Final
try:
    import aionotify  # type: ignore
except (OSError, ModuleNotFoundError):
    aionotify = None  # type: ignore

from opentrons.drivers.smoothie_drivers import driver_3_0
from opentrons.drivers.rpi_drivers import build_gpio_chardev, usb
import opentrons.config
from opentrons.config import pipette_config
from opentrons.config.types import RobotConfig
from opentrons.types import Mount

from .module_control import AttachedModulesControl
from .types import AionotifyEvent, BoardRevision, Axis

if TYPE_CHECKING:
    from opentrons_shared_data.pipette.dev_types import (
        PipetteModel, PipetteName
    )
    from .dev_types import (
        AttachedInstrument, AttachedInstruments,
        InstrumentHardwareConfigs)
    from opentrons.drivers.rpi_drivers.dev_types\
        import GPIODriverLike

MODULE_LOG = logging.getLogger(__name__)


class Controller:
    """ The concrete instance of the controller for actually controlling
    hardware.
    """

    @classmethod
    async def build(
            cls, config: RobotConfig) -> Controller:
        """ Build a Controller instance.

        Use this factory method rather than the initializer to handle proper
        GPIO initialization.

        :param config: A loaded robot config.
        """
        gpio = build_gpio_chardev('gpiochip0')
        gpio.config_by_board_rev()
        await gpio.setup()
        return cls(config, gpio)

    def __init__(self, config: RobotConfig, gpio: GPIODriverLike):
        """ Build a Controller instance.

        Always prefer using :py:meth:`.build` to create an instance of this class. For
        more information on arguments, see that method. If you want to use this
        directly, you must pass in an initialized _and set up_ GPIO driver instance.
        """
        if not opentrons.config.IS_ROBOT:
            MODULE_LOG.warning(
                'This is intended to run on a robot, and while it can connect '
                'to a smoothie via a usb/serial adapter unexpected things '
                'using gpios (such as smoothie reset or light management) '
                'will fail. If you are seeing this message and you are '
                'running on a robot, you need to set the RUNNING_ON_PI '
                'environmental variable to 1.')

        self.config = config or opentrons.config.robot_configs.load()

        self._gpio_chardev: Final = gpio
        self._board_revision: Final = self.gpio_chardev.board_rev
        # We handle our own locks in the hardware controller thank you
        self._smoothie_driver = driver_3_0.SmoothieDriver_3_0_0(
            config=self.config, gpio_chardev=self._gpio_chardev,
            handle_locks=False)
        self._cached_fw_version: Optional[str] = None
        self._usb = usb.USBBus(self._board_revision)
        self._module_controls: Optional[AttachedModulesControl] = None
        try:
            self._event_watcher = self._build_event_watcher()
        except AttributeError:
            MODULE_LOG.warning(
                'Failed to initiate aionotify, cannot watch modules '
                'or door, likely because not running on linux')

    @staticmethod
    def _build_event_watcher():
        watcher = aionotify.Watcher()
        watcher.watch(
            alias='modules',
            path='/dev',
            flags=(aionotify.Flags.CREATE | aionotify.Flags.DELETE))
        return watcher

    @property
    def gpio_chardev(self) -> GPIODriverLike:
        return self._gpio_chardev

    @property
    def board_revision(self) -> BoardRevision:
        return self._board_revision

    @property
    def module_controls(self) -> AttachedModulesControl:
        if not self._module_controls:
            raise AttributeError('Module controls not found.')
        return self._module_controls

    @module_controls.setter
    def module_controls(self, module_controls: AttachedModulesControl):
        self._module_controls = module_controls

    def start_gpio_door_watcher(self, **kargs):
        self.gpio_chardev.start_door_switch_watcher(**kargs)

    def update_position(self) -> Dict[str, float]:
        self._smoothie_driver.update_position()
        return self._smoothie_driver.position

    def move(self, target_position: Dict[str, float],
             home_flagged_axes: bool = True, speed: float = None,
             axis_max_speeds: Dict[str, float] = None):
        with ExitStack() as cmstack:
            if axis_max_speeds:
                cmstack.enter_context(
                    self._smoothie_driver.restore_axis_max_speed(
                        axis_max_speeds))
            self._smoothie_driver.move(
                target_position, home_flagged_axes=home_flagged_axes,
                speed=speed)

    def home(self, axes: List[str] = None) -> Dict[str, float]:
        if axes:
            args: Tuple[Any, ...] = (''.join(axes),)
        else:
            args = tuple()
        return self._smoothie_driver.home(*args)

    def fast_home(
            self, axes: Sequence[str],
            margin: float) -> Dict[str, float]:
        converted_axes = ''.join(axes)
        return self._smoothie_driver.fast_home(converted_axes, margin)

    def _query_mount(
            self,
            mount: Mount,
            expected: Union[PipetteModel, PipetteName, None]
    ) -> AttachedInstrument:
        found_model: Optional[PipetteModel]\
            = self._smoothie_driver.read_pipette_model(  # type: ignore
            mount.name.lower())
        if found_model and found_model not in pipette_config.config_models:
            # TODO: Consider how to handle this error - it bubbles up now
            # and will cause problems at higher levels
            MODULE_LOG.error(
                f'Bad model on {mount.name}: {found_model}')
            found_model = None
        found_id = self._smoothie_driver.read_pipette_id(
            mount.name.lower())

        if found_model:
            config = pipette_config.load(found_model, found_id)
            if expected:
                acceptable = [config.name] + config.back_compat_names
                if expected not in acceptable:
                    raise RuntimeError(f'mount {mount}: instrument'
                                       f' {expected} was requested'
                                       f' but {config.model} is present')
            return {'config': config,
                    'id': found_id}
        else:
            if expected:
                raise RuntimeError(
                    f'mount {mount}: instrument {expected} was'
                    f' requested, but no instrument is present')
            return {'config': None, 'id': None}

    def get_attached_instruments(
            self, expected: Dict[Mount, PipetteName])\
            -> AttachedInstruments:
        """ Find the instruments attached to our mounts.
        :param expected: is ignored, it is just meant to enforce
                          the same interface as the simulator, where
                          required instruments can be manipulated.

        :returns: A dict with mounts as the top-level keys. Each mount value is
            a dict with keys 'model' (containing an instrument model name or
            `None`) and 'id' (containing the serial number of the pipette
            attached to that mount, or `None`). Both mounts will always be
            specified.
        """
        return {mount: self._query_mount(mount, expected.get(mount))
                for mount in Mount}

    def set_active_current(self, axis_currents: Dict[Axis, float]):
        """
        This method sets only the 'active' current, i.e., the current for an
        axis' movement. Smoothie driver automatically resets the current for
        pipette axis to a low current (dwelling current) after each move
        """
        self._smoothie_driver.set_active_current(
            {axis.name: amp for axis, amp in axis_currents.items()})

    @contextmanager
    def save_current(self):
        self._smoothie_driver.push_active_current()
        try:
            yield
        finally:
            self._smoothie_driver.pop_active_current()

    async def _handle_watch_event(self):
        try:
            event = await self._event_watcher.get_event()
        except asyncio.IncompleteReadError:
            MODULE_LOG.debug("incomplete read error when quitting watcher")
            return
        if event is not None:
            if 'ot_module' in event.name:
                event_name = event.name
                flags = aionotify.Flags.parse(event.flags)
                event_description = AionotifyEvent.build(event_name, flags)
                await self.module_controls.handle_module_appearance(event_description)

    async def watch(self, loop: asyncio.AbstractEventLoop):
        can_watch = aionotify is not None
        if can_watch:
            await self._event_watcher.setup(loop)

        while can_watch and (not self._event_watcher.closed):
            await self._handle_watch_event()

    async def connect(self, port: str = None):
        self._smoothie_driver.connect(port)
        await self.update_fw_version()

    @property
    def axis_bounds(self) -> Dict[Axis, Tuple[float, float]]:
        """ The (minimum, maximum) bounds for each axis. """
        return {Axis[ax]: (0, pos) for ax, pos
                in self._smoothie_driver.axis_bounds.items()
                if ax not in 'BC'}

    @property
    def fw_version(self) -> Optional[str]:
        return self._cached_fw_version

    async def update_fw_version(self):
        self._cached_fw_version = self._smoothie_driver.get_fw_version()

    async def update_firmware(self,
                              filename: str,
                              loop: asyncio.AbstractEventLoop,
                              modeset: bool) -> str:
        msg = await self._smoothie_driver.update_firmware(
            filename, loop, modeset)
        await self.update_fw_version()
        return msg

    def engaged_axes(self) -> Dict[str, bool]:
        return self._smoothie_driver.engaged_axes

    def disengage_axes(self, axes: List[str]):
        self._smoothie_driver.disengage_axis(''.join(axes))

    def set_lights(self, button: Optional[bool], rails: Optional[bool]):
        if button is not None:
            self.gpio_chardev.set_button_light(blue=button)
        if rails is not None:
            self.gpio_chardev.set_rail_lights(rails)

    def get_lights(self) -> Dict[str, bool]:
        return {
            'button': self.gpio_chardev.get_button_light()[2],
            'rails': self.gpio_chardev.get_rail_lights()
        }

    def pause(self):
        self._smoothie_driver.pause()

    def resume(self):
        self._smoothie_driver.resume()

    def halt(self):
        self._smoothie_driver.kill()

    def hard_halt(self):
        self._smoothie_driver.hard_halt()

    def probe(self, axis: str, distance: float) -> Dict[str, float]:
        """ Run a probe and return the new position dict
        """
        return self._smoothie_driver.probe_axis(axis, distance)

    def clean_up(self):
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            return
        if hasattr(self, '_event_watcher'):
            if loop.is_running() and self._event_watcher:
                self._event_watcher.close()
        if hasattr(self, 'gpio_chardev'):
            try:
                if not loop.is_closed():
                    self.gpio_chardev.stop_door_switch_watcher(loop)
            except RuntimeError:
                pass

    def configure_mount(self, mount: Mount, config: InstrumentHardwareConfigs):
        mount_axis = Axis.by_mount(mount)
        plunger_axis = Axis.of_plunger(mount)

        self._smoothie_driver.update_steps_per_mm(
            {plunger_axis.name: config['steps_per_mm']})
        self._smoothie_driver.update_pipette_config(
            mount_axis.name, {'home': config['home_pos']})
        self._smoothie_driver.update_pipette_config(
            plunger_axis.name, {'max_travel': config['max_travel']})
        self._smoothie_driver.set_dwelling_current(
            {plunger_axis.name: config['idle_current']})
        ms = config['splits']
        if ms:
            self._smoothie_driver.configure_splits_for(
                {plunger_axis.name: ms})

    def __del__(self):
        self.clean_up()



from __future__ import annotations
import asyncio
import copy
import logging
from threading import Event
from typing import (Dict, Optional, List, Tuple,
                    TYPE_CHECKING, Sequence)
from contextlib import contextmanager

from opentrons_shared_data.pipette import dummy_model_for_name

from opentrons import types
from opentrons.config.pipette_config import (config_models,
                                             config_names,
                                             configs,
                                             load)
from opentrons.config.types import RobotConfig
from opentrons.drivers.smoothie_drivers import SimulatingDriver

from opentrons.drivers.rpi_drivers.gpio_simulator import SimulatingGPIOCharDev
from opentrons.drivers.rpi_drivers import usb_simulator

from . import modules
from .types import BoardRevision, Axis
from .module_control import AttachedModulesControl


if TYPE_CHECKING:
    from opentrons_shared_data.pipette.dev_types import PipetteName
    from .dev_types import (
        AttachedInstrument, AttachedInstruments,
        InstrumentSpec, InstrumentHardwareConfigs)
    from opentrons.drivers.rpi_drivers.dev_types\
        import GPIODriverLike


MODULE_LOG = logging.getLogger(__name__)


class Simulator:
    """ This is a subclass of hardware_control that only simulates the
    hardware actions. It is suitable for use on a dev machine or on
    a robot with no smoothie connected.
    """

    @classmethod
    async def build(
            cls,
            attached_instruments: Dict[types.Mount, Dict[str, Optional[str]]],
            attached_modules: List[str],
            config: RobotConfig,
            loop: asyncio.AbstractEventLoop,
            strict_attached_instruments: bool = True) -> Simulator:
        """ Build the simulator.

        Use this factory method rather than the initializer to handle proper GPIO
        initialization.

        :param attached_instruments: A dictionary describing the instruments
                                     the simulator should consider attached.
                                     If this argument is specified and
                                     :py:meth:`get_attached_instruments` is
                                     called with expectations that do not
                                     match, the call fails. This is useful for
                                     making the simulator match the real
                                     hardware, for instance to check if a
                                     protocol asks for the right instruments.
                                     This dict should map mounts to either
                                     empty dicts or to dicts containing
                                     'model' and 'id' keys.
        :param attached_modules: A list of module model names (e.g.
                                 `'tempdeck'` or `'magdeck'`) representing
                                 modules the simulator should assume are
                                 attached. Like `attached_instruments`, used
                                 to make the simulator match the setup of the
                                 real hardware.
        :param config: The robot config to use
        :param loop: The asyncio event loop to use.
        :param strict_attached_instruments: This param changes the behavior of
                                            the instrument cache. If ``True``,
                                            (default), ``cache_instrument``
                                            calls requesting instruments not
                                            in ``attached_instruments`` will
                                            fail as if the instrument was not
                                            present. If ``False``, those calls
                                            will still pass but give a response
                                            version of 1, while calls
                                            requesting instruments that _are_
                                            present get the full number.
        """
        gpio = SimulatingGPIOCharDev('gpiochip0')
        await gpio.setup()
        return cls(
            attached_instruments,
            attached_modules,
            config,
            loop,
            gpio,
            strict_attached_instruments)

    def __init__(
            self,
            attached_instruments: Dict[types.Mount, Dict[str, Optional[str]]],
            attached_modules: List[str],
            config: RobotConfig, loop: asyncio.AbstractEventLoop,
            gpio_chardev: GPIODriverLike,
            strict_attached_instruments: bool = True) -> None:
        """ Initialize the simulator.

        Always prefer using :py:meth:`.build` to create an instance of this class. For
        more information on arguments, see that method. If you want to use this
        directly, you must pass in an initialized _and set up_ GPIO driver instance.
        """
        self.config = config
        self._loop = loop
        self._smoothie_driver = SimulatingDriver()
        self._gpio_chardev = gpio_chardev

        def _sanitize_attached_instrument(
                passed_ai: Dict[str, Optional[str]] = None)\
                -> InstrumentSpec:
            if not passed_ai or not passed_ai.get('model'):
                return {'model': None, 'id': None}
            if passed_ai['model'] in config_models:
                return passed_ai  # type: ignore
            if passed_ai['model'] in config_names:
                return {
                    'model':
                    dummy_model_for_name(passed_ai['model']),  # type: ignore
                    'id': passed_ai.get('id')}
            raise KeyError(
                'If you specify attached_instruments, the model '
                'should be pipette names or pipette models, but '
                f'{passed_ai["model"]} is not')

        self._attached_instruments = {
            m: _sanitize_attached_instrument(attached_instruments.get(m))
            for m in types.Mount}
        self._stubbed_attached_modules = attached_modules
        self._position = copy.copy(self._smoothie_driver.homed_position)
        # Engaged axes start all true in smoothie for some reason so we
        # imitate that here
        # TODO(LC2642019) Create a simulating driver for smoothie instead of
        # using a flag

        self._engaged_axes = {ax: True for ax in self._smoothie_driver.homed_position}
        self._lights = {'button': False, 'rails': False}
        self._run_flag = Event()
        self._run_flag.set()
        self._log = MODULE_LOG.getChild(repr(self))
        self._strict_attached = bool(strict_attached_instruments)
        self._board_revision = BoardRevision.OG
        self._usb = usb_simulator.USBBusSimulator(self._board_revision)
        # TODO (lc 05-12-2021) In a follow-up refactor that pulls the execution
        # manager responsbility into the controller/backend itself as opposed
        # to the hardware api controller.
        self._module_controls: Optional[AttachedModulesControl] = None

    @property
    def gpio_chardev(self) -> GPIODriverLike:
        return self._gpio_chardev

    @property
    def module_controls(self) -> AttachedModulesControl:
        if not self._module_controls:
            raise AttributeError('Module controls not found.')
        return self._module_controls

    @module_controls.setter
    def module_controls(self, module_controls: AttachedModulesControl):
        self._module_controls = module_controls

    def update_position(self) -> Dict[str, float]:
        return self._position

    def move(self, target_position: Dict[str, float],
             home_flagged_axes: bool = True, speed: float = None,
             axis_max_speeds: Dict[str, float] = None):
        self._position.update(target_position)
        self._engaged_axes.update({ax: True
                                   for ax in target_position})

    def home(self, axes: List[str] = None) -> Dict[str, float]:
        # driver_3_0-> HOMED_POSITION
        checked_axes = axes or 'XYZABC'
        self._position.update({ax: self._smoothie_driver.homed_position[ax]
                               for ax in checked_axes})
        self._engaged_axes.update({ax: True
                                   for ax in checked_axes})
        return self._position

    def fast_home(
            self, axis: Sequence[str], margin: float) -> Dict[str, float]:
        for ax in axis:
            self._position[ax] = self._smoothie_driver.homed_position[ax]
            self._engaged_axes[ax] = True
        return self._position

    def _attached_to_mount(
            self,
            mount: types.Mount,
            expected_instr: Optional[PipetteName]
    ) -> AttachedInstrument:
        init_instr = self._attached_instruments.get(
            mount,
            {'model': None, 'id': None})
        found_model = init_instr['model']
        back_compat: List['PipetteName'] = []
        if found_model:
            back_compat = configs[found_model].get('backCompatNames', [])
        if expected_instr and found_model\
                and (not found_model.startswith(expected_instr)
                     and expected_instr not in back_compat):
            if self._strict_attached:
                raise RuntimeError(
                    'mount {}: expected instrument {} but got {}'
                    .format(mount.name, expected_instr, found_model))
            else:
                return {
                    'config': load(dummy_model_for_name(expected_instr)),
                    'id': None}
        elif found_model and expected_instr:
            # Instrument detected matches instrument expected (note:
            # "instrument detected" means passed as an argument to the
            # constructor of this class)
            return {'config': load(found_model, init_instr['id']),
                    'id': init_instr['id']}
        elif found_model:
            # Instrument detected and no expected instrument specified
            return {'config': load(found_model, init_instr['id']),
                    'id': init_instr['id']}
        elif expected_instr:
            # Expected instrument specified and no instrument detected
            return {
                'config': load(dummy_model_for_name(expected_instr)),
                'id': None}
        else:
            # No instrument detected or expected
            return {
                'config': None,
                'id': None}

    def get_attached_instruments(
            self, expected: Dict[types.Mount, PipetteName]
    ) -> AttachedInstruments:
        """ Update the internal cache of attached instruments.

        This method allows after-init-time specification of attached simulated
        instruments. The method will return
        - the instruments specified at init-time, or if those do not exists,
        - the instruments specified in expected, or if that is not passed,
        - nothing

        :param expected: A mapping of mount to instrument model prefixes. When
                         loading instruments from a prefix, we return the
                         lexically-first model that matches the prefix. If the
                         models specified in expected do not match the models
                         specified in the `attached_instruments` argument of
                         :py:meth:`__init__`, :py:attr:`RuntimeError` is
                         raised.
        :raises RuntimeError: If an instrument is expected but not found.
        :returns: A dict of mount to either instrument model names or `None`.
        """
        return {
            mount: self._attached_to_mount(mount, expected.get(mount))
            for mount in types.Mount
        }

    def set_active_current(self, axis_currents: Dict[Axis, float]):
        pass

    async def watch(self):
        new_mods_at_ports = [
            modules.ModuleAtPort(
                port=f'/dev/ot_module_sim_{mod}{str(idx)}', name=mod)
            for idx, mod
            in enumerate(self._stubbed_attached_modules)]
        await self.module_controls.register_modules(
            new_mods_at_ports=new_mods_at_ports)

    @contextmanager
    def save_current(self):
        yield

    @property
    def axis_bounds(self) -> Dict[Axis, Tuple[float, float]]:
        """ The (minimum, maximum) bounds for each axis. """
        return {Axis[ax]: (0, pos) for ax, pos
                in self._smoothie_driver.axis_bounds.items()
                if ax not in 'BC'}

    @property
    def fw_version(self) -> Optional[str]:
        return 'Virtual Smoothie'

    async def update_fw_version(self):
        pass

    @property
    def board_revision(self) -> BoardRevision:
        return self._board_revision

    async def update_firmware(self, filename, loop, modeset) -> str:
        return 'Did nothing (simulating)'

    def engaged_axes(self):
        return self._engaged_axes

    def disengage_axes(self, axes: List[str]):
        self._engaged_axes.update({ax: False for ax in axes})

    def set_lights(self, button: Optional[bool], rails: Optional[bool]):
        if button is not None:
            self._lights['button'] = button
        if rails is not None:
            self._lights['rails'] = rails

    def get_lights(self) -> Dict[str, bool]:
        return self._lights

    def pause(self):
        self._run_flag.clear()

    def resume(self):
        self._run_flag.set()

    def halt(self):
        self._run_flag.set()

    def hard_halt(self):
        self._run_flag.set()

    def probe(self, axis: str, distance: float) -> Dict[str, float]:
        self._position[axis.upper()] = self._position[axis.upper()] + distance
        return self._position

    def clean_up(self):
        pass

    def configure_mount(
            self, mount: types.Mount, config: InstrumentHardwareConfigs):
        mount_axis = Axis.by_mount(mount)
        plunger_axis = Axis.of_plunger(mount)

        self._smoothie_driver.update_steps_per_mm(
            {plunger_axis.name: config['steps_per_mm']})
        self._smoothie_driver.update_pipette_config(
            mount_axis.name, {'home': config['home_pos']})
        self._smoothie_driver.update_pipette_config(
            plunger_axis.name, {'max_travel': config['max_travel']})
        self._smoothie_driver.set_dwelling_current(
            {plunger_axis.name: config['idle_current']})
        ms = config['splits']
        if ms:
            self._smoothie_driver.configure_splits_for(
                {plunger_axis.name: ms})
